#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "ilgate.h"

#define LOGFILE "/var/log/auth.srv"

void
usage ()
{
	fprintf (stderr, "usage: auth.srv [-d]\n");
	exit (1);
}

struct task {
	struct task *next;
	struct task *prev;
	int (*func)(struct task *, struct ilconn *);
	struct ilconn *ilconn;
	int pc;
	void *userdata;
};

struct task tasks;

void
make_task (int (*func)(struct task *, struct ilconn *ilconn), void *ilconn)
{
	struct task *tp;

	tp = calloc (1, sizeof *tp);
	tp->func = func;
	tp->ilconn = ilconn;
	tp->next = &tasks;
	tp->prev = tasks.prev;
	tasks.prev->next = tp;
	tasks.prev = tp;
}

void *
task_userdata (struct task *tp, int size)
{
	if (tp->userdata == NULL)
		tp->userdata = calloc (1, size);
	return (tp->userdata);
}

void
task_exit (struct task *tp)
{
	if (tp->userdata)
		free (tp->userdata);
	tp->next->prev = tp->prev;
	tp->prev->next = tp->next;
	free (tp);
}

int background_flag;

int
main (int argc, char **argv)
{
	int s;
	struct task *tp, *nexttp;
	fd_set rset;
	fd_set wset;
	fd_set eset;
	int n;
	int addrlen;
	struct sockaddr_in addr;
	char buf[10000];
	struct ip *ip;
	int iphdrlen;
	int c;

	if (geteuid () != 0) {
		fprintf (stderr, "must run as root\n");
		exit (1);
	}

	while ((c = getopt (argc, argv, "db")) != EOF) {
		switch (c) {
		case 'd':
			debug_flag = 1;
			break;
		case 'b':
			background_flag = 1;
			break;
		default:
			usage ();
		}
	}

	if (background_flag) {
		int pid;

		if ((pid = fork ()) < 0) {
			perror ("fork");
			exit (1);
		}

		if (pid != 0)
			exit (0);

		if (freopen (LOGFILE, "a", stdout) == NULL) {
			fprintf (stderr, "can't open logfile %s\n", LOGFILE);
			exit (1);
		}

		/* 
		 * I tried to avoid logging any secrets, but just in case,
		 * limit the access to the log file.
		 */
		chown (LOGFILE, 0, 0);
		chmod (LOGFILE, 0600);
		
		dup2 (1, 2);
	}

	tasks.next = &tasks;
	tasks.prev = &tasks;

	if ((s = socket (AF_INET, SOCK_RAW, 40)) < 0) {
		printf ("setting up SOCK_RAW: %s", strerror (errno));
		exit (1);
	}
	raw_sock = s;

	fcntl (s, F_SETFL, O_NONBLOCK);

	while (1) {
		for (tp = tasks.next, nexttp = tp->next;
		     tp != &tasks;
		     tp = nexttp, nexttp = tp->next) {
			if ((*tp->func)(tp, tp->ilconn) < 0) {
				il_abort (tp->ilconn);
				task_exit (tp);
			}
		}

		FD_ZERO (&rset);
		FD_ZERO (&wset);
		FD_ZERO (&eset);

		FD_SET (s, &rset);
		FD_SET (s, &eset);

		n = select (s + 1, &rset, &wset, &eset, NULL);

		if (n < 0) {
			printf ("select: %s\n", strerror (errno));
			exit (1);
		}

		if (FD_ISSET (s, &rset)) {
			addrlen = sizeof addr;
			n = recvfrom (s, buf, sizeof buf, 0,
				      (struct sockaddr *)&addr, &addrlen);
			if (n < 0) {
				printf ("recvfrom error: %s\n",
					strerror (errno));
				continue;
			}

			ip = (struct ip *)buf;
			
			if (ip->ip_p != 40)
				continue;

			iphdrlen = ip->ip_hl * 4;
			
			if (iphdrlen  > n) {
				printf ("ip pkt too small\n");
				continue;
			}

			ilrcv (ip->ip_src, ip->ip_dst,
			       buf + iphdrlen, n - iphdrlen);
		}

		/* watch out for contuines above */
	}
}

int
task_pc (struct task *tp)
{
	return (tp->pc);
}

void
task_set_pc (struct task *tp, int pc)
{
	tp->pc = pc;
}

